class RTFMException(Exception):
    def __init__(self, msg: str, link: str) -> None:
        super().__init__(msg)
        self.link = link
        self.msg = msg

    def __str__(self) -> str:
        return f"\n{self.msg}\nSee {self.link} for more information."



from py4j.java_gateway import JavaObject
from typing import Optional

from ..transformations_dag import Metadata, TransformationType
from .transformation_node_builder import TransformationNodeBuilder

from ..py4j_utils import iterate_java_object


class LogicalRelationNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["type"] = "relation"
        metadata["file_format"] = node.relation().fileFormat().toString()

        if node.relation().getClass().getSimpleName() == "HadoopFsRelation":
            metadata["location_files"] = [
                path.toString() for path in
                iterate_java_object(node.relation().location().rootPaths())
            ]
            metadata["partition_schema"] = node.relation().location().partitionSchema().toString()

        catalog_table_option = node.catalogTable()
        if not catalog_table_option.isEmpty():
            catalog_table = catalog_table_option.get()
            metadata["database"] = catalog_table.database()
            metadata["table"] = catalog_table.identifier().table()
            metadata["partition_columns"] = [name for name in iterate_java_object(catalog_table.partitionColumnNames())]
            metadata["storage"] = catalog_table.storage().toString()

    def _get_type(self) -> TransformationType:
        return TransformationType.Relation

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 0



from typing import Dict, Set

from ..transformations_dag import Metadata, TransformationColumn, TransformationNode, TransformationType

from .dag_simplifier import DagSimplifier


class MergeDatasourceAndAlias(DagSimplifier):
    """
    When the PySpark dataframe represents a named table, it generates
    two nodes: one containing the table data itself, and another one
    containing an alias. This simplifier unifies both adding an
    attribute to the table node.
    When the datasource is a logical RDD, the alias is not added, so
    this simplifier considers that case and skips it.
    """

    def __init__(self) -> None:
        self.already_merged_relations: Set[TransformationNode] = set()

    def can_apply(self, df: TransformationNode) -> bool:
        if df.type != TransformationType.Alias:
            return False
        alias_node = df
        if alias_node in self.already_merged_relations:
            return False
        if len(alias_node.children) != 1:
            return False
        relation_node = alias_node.children[0]
        if relation_node.type != TransformationType.Relation:
            return False
        return True

    def apply(self, original: TransformationNode) -> TransformationNode:
        alias_node = original
        relation_node = original.children[0]
        new_node = TransformationNode(
            type=TransformationType.Relation,
            metadata=relation_node.metadata,
            children=relation_node.children,
            columns=self._remove_links(alias_node.columns),
        )
        for c in new_node.columns.values():
            c.node_id = new_node.id
        self.already_merged_relations.add(new_node)
        return new_node

    def _remove_links(self, alias_cols: Dict[int, TransformationColumn]) -> Dict[int, TransformationColumn]:
        for col in alias_cols.values():
            col.links = []
        return alias_cols



import dataclasses
from collections import defaultdict
from typing import Generator, Tuple, Optional, List, Dict
from .transformations_dag import TransformationNode


@dataclasses.dataclass
class DFSNodeData:
    # identifier of the parent node
    # Note that, for the root node only, the `parent_id` is None
    parent_id: Optional[int]

    # parent node reference
    parent: Optional[TransformationNode]

    # the depth of the node in the DFS
    depth: int


def dfs(root: TransformationNode) -> Generator[Tuple[DFSNodeData, TransformationNode], None, None]:
    """Returns a generator that iterates over the tree in depth-first order.
    Each element yielded by the generator is a tuple of (DFSNodeData, node)."""

    parents: Dict[int, Optional[int]] = {root.id: None}
    nodes: Dict[int, TransformationNode] = {root.id: root}
    stack: List[int] = [root.id]
    depths: Dict[int, int] = {root.id: 0}

    while stack:
        node_id = stack.pop()

        # yield the node data and the node itself
        parent_id = parents[node_id]
        data = DFSNodeData(
            parent_id=parents[node_id],
            parent=nodes[parent_id] if parent_id is not None else None,
            depth=depths[node_id],
        )
        yield data, nodes[node_id]

        # push all children to the stack in order to iterate them later
        # we assume this is a tree, so no cycles should happen here
        for child in nodes[node_id].children:
            nodes[child.id] = child
            parents[child.id] = node_id
            depths[child.id] = depths[node_id] + 1

            stack.append(child.id)



from typing import Dict, List, Set, TypeVar

from ..transformations_dag import Metadata, TransformationColumn, TransformationNode, TransformationType, JoinCondition

from .dag_simplifier import DagSimplifier


class MergeJoinAndProject(DagSimplifier):
    """
    When the PySpark dataframe has a join, the resulting DAG has
    two nodes: one is a Join, and the other is a Project. With this simplification
    both nodes are merged.
    When joining by column name, the node has one column representing the
    column coming from the "left" df, and another for the "right" df. Our
    resulting merged join node contains only one column that has links
    to the two origin columns.
    """

    def __init__(self) -> None:
        self.already_merged_joins: Set[TransformationNode] = set()

    def can_apply(self, dag: TransformationNode) -> bool:
        if dag.type != TransformationType.Project:
            return False
        project_node = dag
        if len(project_node.children) != 1:
            return False
        join_node = project_node.children[0]
        if join_node.type != TransformationType.Join:
            return False
        if join_node.metadata["join_type"] == "Cross":
            # Cross Join doesn't add the extra Project, so removing it would be wrong
            return False
        if join_node in self.already_merged_joins:
            # Avoid collapsing all subsequent Projects
            return False
        condition = join_node.metadata["condition"]
        if not condition.is_equi_join:
            return False
        project_column_names = [c.name for c in project_node.columns.values()]
        join_column_names = [c.name for c in join_node.columns.values()]
        collapsed_col_names = [n for n in condition.equi_join_columns.keys()]
        return sorted(project_column_names + collapsed_col_names) == sorted(join_column_names)

    def apply(self, original: TransformationNode) -> TransformationNode:
        project_node = original
        join_node = original.children[0]
        new_node = TransformationNode(
            type=TransformationType.Join,
            metadata=self._merge_metadata(project_node.metadata, join_node.metadata),
            children=join_node.children,
            columns=self._merge_columns(project_node.columns, join_node.columns, join_node.metadata["condition"]),
        )
        for c in new_node.columns.values():
            c.node_id = new_node.id
        self.already_merged_joins.add(new_node)
        return new_node

    def _merge_metadata(self, proj_meta: Metadata, join_meta: Metadata) -> Metadata:
        return {
            "condition": join_meta["condition"],
            "join_type": join_meta["join_type"],
            "schema_string": proj_meta["schema_string"],
        }

    def _merge_columns(self, pcols: Dict[int, TransformationColumn], jcols: Dict[int, TransformationColumn], cond: JoinCondition) -> Dict[int, TransformationColumn]:
        for col in pcols.values():
            assert len(col.links) == 1
            link = col.links[0]

            links = []
            tree_string = ""
            if link.name in cond.equi_join_columns.keys():
                links = flatten([jcols[i].links for i in cond.equi_join_columns[link.name]])
                tree_string = cond.tree_string
            else:
                links = jcols[link.id].links
                tree_string = col.tree_string

            col.links = links
            col.tree_string = tree_string

        return pcols


T = TypeVar("T")
def flatten(l: List[List[T]]) -> List[T]:
    return [item for sublist in l for item in sublist]



from py4j.java_gateway import JavaObject
from typing import Optional

from ..transformations_dag import Metadata, TransformationType
from .transformation_node_builder import TransformationNodeBuilder


class UnknownNodeBuilder(TransformationNodeBuilder):
    """This class is used to build nodes for transformations that are not supported.
    Instead of failing and leaving the users without any visualization, we can opt
    to return an "unknown" transformation and at least provide some kind of output.
    """

    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["inferred_type"] = node.getClass().getSimpleName()
        # toString() is a method of the `TreeNode` class, so it should be available in all nodes
        metadata["representation"] = node.toString()

    def _get_type(self) -> TransformationType:
        return TransformationType.Unknown

    def _expected_number_of_nodes(self) -> Optional[int]:
        return None



from py4j.java_gateway import JavaObject
from pyspark.sql import DataFrame
from typing import List, Dict, Any, Generator, Optional

from .dag_simplification import simplify, default_simplifiers
from .dag_simplification.dag_simplifier import DagSimplifier
from .transformations_dag import TransformationNode
from .transformation_node_builders import build_dag_from_java_object
from .transformation_node_builders.repository import Repository as TnbRepository, create_default
from .py4j_utils import iterate_java_object


def build_dag(df: DataFrame, dag_simplifiers: List[DagSimplifier] = default_simplifiers(), tnb_repository: TnbRepository=create_default()) -> TransformationNode:
    """
    Builds a DAG from a PySpark's DataFrame.
    
    :param    df:               The PySpark's DataFrame.
    :param    dag_simplifiers:  The list of DAG simplifiers to be applied.
    :param    tnb_repository:   A repository of specific transformation node builders.
    :returns: The DAG for the passed PySpark's DataFrame.
    """
    first_child = _get_last_transformation(df)
    raw_dag = build_dag_from_java_object(node=first_child, repository=tnb_repository)
    return simplify(raw_dag, dag_simplifiers)


def _get_last_transformation(df: DataFrame) -> JavaObject:
    # we add a dummy projection of all columns to Spark resolves the next
    # transformations in the plan. Otherwise, the columns will be "unresolved"
    df = df.select("*")
    root = df._jdf.queryExecution().logical()

    # skip the 'Project [*]' node
    return next(iterate_java_object(root.children()))



from py4j.java_gateway import JavaObject
from typing import Optional

from ..transformations_dag import Metadata, TransformationType
from .transformation_node_builder import TransformationNodeBuilder

from ..py4j_utils import iterate_java_object


class DeduplicateNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["deduplicate_columns"] = [col.sql() for col in iterate_java_object(node.references())]

    def _get_type(self) -> TransformationType:
        return TransformationType.Deduplicate

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1



"""
Defines the class that allows spark-board users to set the default values for
the website settings.
"""

import dataclasses


@dataclasses.dataclass
class DefaultSettings:
    """Default settings for the HTML DAG renderer. The settings can be changed through the UI, but
    here we only define the default value they'll take when the website loads."""

    # enable or disable all "bouncy" node animations
    animationEnabled: bool = True

    # nodes are animated (pull linked nodes) when dragged with the mouse on the screen
    animationEnabledOnDrag: bool = True

    # Data source nodes will appear multiple times on each used transformation that
    # accesses it.
    duplicateDataSources: bool = False

    # whenever a node's size is changed, all the DAG layout is reacommodated (but still
    # respecting the nodes that were manually placed).
    organizeNodesOnResize: bool = False

    # inverts the DAG layout (from data source nodes to last transformation or the
    # other way around).
    invertDag: bool = True



from pyspark.sql.session import SparkSession
from spark_board.html import dump_dataframe, DefaultSettings
from pathlib import Path


def test_spark_board(spark: SparkSession, tmp_path: Path) -> None:
    """This test intends to be a very simple smoke test for the normal spark-board
    usage."""

    # create a data frame
    df = spark.createDataFrame([], schema="struct<a:double, b:double, c:double>")
    df = df.withColumn("sum", (df.a + df.b) / df.c)

    OUT_DIR = tmp_path/"spark_board_output"

    # dump it expecting no errors
    dump_dataframe(
        df=df,
        output_dir=str(OUT_DIR),
        overwrite=True,  # overwrite output_dir if it already exists
        default_settings=DefaultSettings(),  # override default settings if desired
    )

    assert (OUT_DIR/"model.js").exists()



from ..transformations_dag import TransformationNode

from .dag_simplifier import DagSimplifier
from .merge_join_and_project import MergeJoinAndProject
from .merge_datasource_and_alias import MergeDatasourceAndAlias

from typing import Dict, List, Set, TypeVar


def default_simplifiers() -> List[DagSimplifier]:
    """
    Returns the list of default simplifiers implemented within spark_board.
    """
    return [
        MergeJoinAndProject(),
        MergeDatasourceAndAlias(),
    ]


def simplify(raw_dag: TransformationNode, simplifiers: List[DagSimplifier]=default_simplifiers()) -> TransformationNode:
    """
    Applies the passed simplifiers to the passed DAG, with a "child-first" policy:
    the simplifiers ar applied to the leaves of the tree first, so when a simplifier
    is being applied to a node, the children are already simplified.

    :param      raw_dag:      The DAG to be simplified.
    :param      simplifiers:  The simplifiers to apply.

    :returns:   The simplified DAG.
    """
    raw_dag.children = [simplify(child) for child in raw_dag.children]

    result_dag = raw_dag
    for simplifier in simplifiers:
        if simplifier.can_apply(result_dag):
            result_dag = simplifier.apply(result_dag)
    return result_dag



from py4j.java_gateway import JavaObject
from typing import Optional

from ..transformations_dag import Metadata, TransformationType
from .transformation_node_builder import TransformationNodeBuilder

from ..py4j_utils import iterate_java_object


class RepartitionNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["num_partitions"] = node.numPartitions()
        metadata["expressions"] = [exp.sql() for exp in iterate_java_object(node.expressions())]

    def _get_type(self) -> TransformationType:
        return TransformationType.Repartition

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1



from py4j.java_gateway import JavaObject
from typing import List, Dict, Any, Generator, Optional, Tuple

from ..transformations_dag import JoinCondition, Metadata, TransformationColumn, TransformationNode, TransformationType

from .repository import Repository, create_default


def build_dag_from_java_object(node: JavaObject, repository: Repository=create_default()) -> TransformationNode:
    """
    Builds a DAG from the PySpark DataFrame's root java object.
    
    :param      node:                           The root java node of the dataframe to be transformed into a DAG.
    :param      repository:                     A repository of specific transformation node builders.
    :returns:   The DAG representing the DataFrame.
    """
    builder = repository.get_builder(node.nodeName())
    return builder.build(node, lambda java_node: build_dag_from_java_object(java_node, repository))



from py4j.java_gateway import JavaObject
from typing import Any, Dict, List, Optional

from ..transformations_dag import Metadata, TransformationType
from .transformation_node_builder import TransformationNodeBuilder

from ..py4j_utils import iterate_java_object


class SortNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["order"] = self._order_as_dict(node.order())

    def _get_type(self) -> TransformationType:
        return TransformationType.Sort

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1

    def _order_as_dict(self, java_order: JavaObject) -> Dict[str, Any]:
        criterias = [o for o in iterate_java_object(java_order)]
        if len(criterias) == 1:
            direction = 'Ascending' if criterias[0].isAscending() else 'Descending'
            column_names = [f"'{ref.sql()}'" for ref in iterate_java_object(criterias[0].references())]
            simple_str = f"{direction} by {', '.join(column_names)}"

            return {
                "SQLs": [criterias[0].sql()],
                "simple_str": simple_str,
            }
        else:
            return {
                "SQLs": [c.sql() for c in criterias],
                "simple_str": "Multiple Columns"
            }



import os


def allow_unknown_transformations() -> bool:
    """If "true", the graph will include nodes for transformations that are not
    recognized by spark-board. If "false", unkonwn transformations will raise
    errors. We may want to disable this in testing environments to find unsupported
    transformations, but in production we probably want to show the visualization
    anyway (whith possibly a couple of unkown transformations with a default
    representation).
    """

    return _get_boolean("SPARK_BOARD_ALLOW_UNKNOWN_TRANSFORMATIONS", default=True)


def _get_boolean(name: str, default: bool) -> bool:
    """Read a boolean value from the environment variable `name`. If the value is not set,
    returns the `default` value."""

    value = os.environ.get(name)
    if value is None:
        return default

    return value.lower() in ("true", "1", "yes")



from py4j.java_gateway import JavaObject
from typing import Dict, List, Optional, Tuple

from ..transformations_dag import JoinCondition, Metadata, TransformationType
from .transformation_node_builder import TransformationNodeBuilder

from ..py4j_utils import iterate_java_object


class JoinNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["join_type"] = node.joinType().toString()
        metadata["condition"] = self._extract_condition(node)

    def _get_type(self) -> TransformationType:
        return TransformationType.Join

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 2

    def _extract_condition(self, node: JavaObject) -> JoinCondition:
        if node.condition().isEmpty():
            return JoinCondition("", "", False, {})
        cond = node.condition().get()
        is_equi_join, equi_join_columns = self._is_equi_join(cond)
        reference_ids = [ref.exprId().id() for ref in iterate_java_object(cond.references())]
        return JoinCondition(cond.sql(), cond.treeString(), is_equi_join, equi_join_columns)

    def _get_column_name(self, java_object: JavaObject) -> Optional[str]:
        if java_object.nodeName() == "AttributeReference":
            return str(java_object.name())
        elif java_object.nodeName() == "Cast":
            return self._get_column_name(java_object.child())
        else:
            return None

    def _is_equi_join(self, cond: JavaObject) -> Tuple[bool, Dict[str, List[int]]]:
        node_name = cond.nodeName()
        if node_name == "EqualTo":
            cols = [col for col in iterate_java_object(cond.children())]
            reference_ids = [ref.exprId().id() for ref in iterate_java_object(cond.references())]
            if len(cols) != 2:
                return False, {}
            name1 = self._get_column_name(cols[0])
            name2 = self._get_column_name(cols[1])
            if name1 is not None and name1 == name2:
                return True, {name1: reference_ids}
            else:
                return False, {}

        if node_name == "And":
            equi_join_columns = {}
            for child in iterate_java_object(cond.children()):
                child_is_equi, child_cols = self._is_equi_join(child)
                if child_is_equi:
                    equi_join_columns.update(child_cols)
                else:
                    return False, {}
            return True, equi_join_columns

        return False, {}



from typing import Callable, Dict, Optional

from .aggregate_node_builder import AggregateNodeBuilder
from .alias_node_builder import AliasNodeBuilder
from .deduplicate_node_builder import DeduplicateNodeBuilder
from .except_node_builder import ExceptNodeBuilder, ExceptAllNodeBuilder
from .expand_node_builder import ExpandNodeBuilder
from .filter_node_builder import FilterNodeBuilder
from .generate_node_builder import GenerateNodeBuilder
from .intersect_node_builder import IntersectNodeBuilder, IntersectAllNodeBuilder
from .join_node_builder import JoinNodeBuilder
from .limit_node_builder import LimitNodeBuilder
from .logical_rdd_node_builder import LogicalRDDNodeBuilder
from .logical_relation_node_builder import LogicalRelationNodeBuilder
from .project_node_builder import ProjectNodeBuilder
from .repartition_node_builder import RepartitionNodeBuilder
from .sample_node_builder import SampleNodeBuilder
from .sort_node_builder import SortNodeBuilder
from .transformation_node_builder import TransformationNodeBuilder
from .union_node_builder import UnionNodeBuilder
from .unknown_node_builder import UnknownNodeBuilder
from .window_node_builder import WindowNodeBuilder


def _default_node_builders() -> Dict[str, 'TransformationNodeBuilder']:
    """
    Returns the default set of objects that can parse each transformation sub-tree.
    """
    return {
        "SubqueryAlias": AliasNodeBuilder(),
        "Project": ProjectNodeBuilder(),
        "Filter": FilterNodeBuilder(),
        "LogicalRDD": LogicalRDDNodeBuilder(),
        "LogicalRelation": LogicalRelationNodeBuilder(),
        "Join": JoinNodeBuilder(),
        "Generate": GenerateNodeBuilder(),
        "Aggregate": AggregateNodeBuilder(),
        "Window": WindowNodeBuilder(),
        "Sort": SortNodeBuilder(),
        "Union": UnionNodeBuilder(),
        "GlobalLimit": LimitNodeBuilder(),
        "LocalLimit": LimitNodeBuilder(),
        "Repartition": RepartitionNodeBuilder(),
        "Deduplicate": DeduplicateNodeBuilder(),
        "Except": ExceptNodeBuilder(),
        "Except All": ExceptAllNodeBuilder(),
        "Intersect": IntersectNodeBuilder(),
        "Intersect All": IntersectAllNodeBuilder(),
        "Sample": SampleNodeBuilder(),
        "RepartitionByExpression": RepartitionNodeBuilder(),
        "Expand": ExpandNodeBuilder(),
    }


class Repository(object):
    """
    This class describes a repository of 'TransformationNodeBuilder's indexed by name.
    The repository is dynamic: builders can be added/removed from the repository.
    If an unknown transformation name is requested, it will return an UnknownNodeBuilder by default,
    but this behavior can be changed by using the method 'with_default_builder'.
    """

    def __init__(self, initial_node_builders: Dict[str, TransformationNodeBuilder] = {}, allow_unknown_transformations: bool=True) -> None:
        """
        Constructs a new instance.
        :param      initial_node_builders:          The initial node builders
        :param      allow_unknown_transformations:  Configures the behavior if an unknown transformation node name is found.
                                                    If True, a syntetic node will be added to the DAG;
                                                    If False, an error will be thrown.
        """
        self.transformation_builders: Dict[str, TransformationNodeBuilder] = initial_node_builders
        self.default_node_builder_provider: Callable[[str], TransformationNodeBuilder] = self._print_warning_and_return_unknown_builder

    def _print_warning_and_return_unknown_builder(self, name: str) -> TransformationNodeBuilder:
        print(f"[WARN] Unknown node name: '{name}'")
        return UnknownNodeBuilder()

    def get_builder(self, node_name: str) -> TransformationNodeBuilder:
        builder = self.transformation_builders.get(node_name)
        if not builder:
            builder = self.default_node_builder_provider(node_name)
        return builder

    def add_builder(self, node_name: str, builder: TransformationNodeBuilder) -> 'Repository':
        self.transformation_builders[node_name] = builder
        return self

    def remove_builder(self, node_name: str) -> 'Repository':
        self.transformation_builders.pop(node_name)
        return self

    def with_default_builder(self, provider: Callable[[str], TransformationNodeBuilder]) -> 'Repository':
        """This method sets the callback that will be called whenever a node whose transformation
        type is not known is encountered. This allows defining how those nodes will be handled (which
        information will be shown/extracted from it) or whether to fail, for example."""
        self.default_node_builder_provider = provider
        return self


def _raise_error_on_unknown_node(unknown_node_name: str) -> TransformationNodeBuilder:
    raise NotImplementedError(f"Transformation not supported: '{unknown_node_name}'")


def create_empty() -> Repository:
    """
    Creates a 'Repository' without any builder.
    If a builder is requested, it will return an 'UnknownNodeBuilder' by default.
    """
    return Repository()


def create_default() -> Repository:
    """
    Creates a 'Repository' with the default set of 'TransformationNodeBuilder's.
    If a builder with an unknown name is requested, it will return an 'UnknownNodeBuilder' by default.
    """
    return Repository(_default_node_builders())


def create_strict() -> Repository:
    """
    Creates a 'Repository' with the default set of 'TransformationNodeBuilder's.
    If a builder with an unknown name is requested, it will raise a 'NotImplementedError'.
    """
    return create_default().with_default_builder(_raise_error_on_unknown_node)



from py4j.java_gateway import JavaObject
from typing import Optional

from ..transformations_dag import Metadata, TransformationType
from .transformation_node_builder import TransformationNodeBuilder

from ..py4j_utils import iterate_java_object


class AggregateNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        aggregate_expressions = []
        for aggregate_expression in iterate_java_object(node.aggregateExpressions()):
            aggregate_expressions.append(aggregate_expression.sql())

        grouping_expressions = []
        for grouping_expression in iterate_java_object(node.groupingExpressions()):
            grouping_expressions.append(grouping_expression.sql())

        metadata['aggregate_expressions'] = aggregate_expressions
        metadata['grouping_expressions'] = grouping_expressions
        if len(grouping_expressions) == 0:
            metadata['grouping_column'] = "No Grouping Column"
        elif len(grouping_expressions) <= 2:
            metadata['grouping_column'] = ', '.join(grouping_expressions)
        else:
            metadata['grouping_column'] = "Multiple Columns"

    def _get_type(self) -> TransformationType:
        return TransformationType.Aggregate

    def _get_columns(self, java_node: JavaObject) -> JavaObject:
        return java_node.aggregateExpressions()

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1



from py4j.java_gateway import JavaObject
from typing import List, Dict, Optional, Callable

from ..transformations_dag import Metadata, TransformationColumn, TransformationNode, TransformationType
from ..py4j_utils import iterate_java_object


class TransformationNodeBuilder(object):

    def build(self, java_node: JavaObject, on_node: Callable[[JavaObject], TransformationNode]) -> TransformationNode:
        if self._expected_number_of_nodes() is not None:
            assert java_node.children().size() == self._expected_number_of_nodes(), java_node.children().size()

        metadata: Dict[str, str] = {}
        self._extract_common_metadata(java_node, metadata)
        self._extract_metadata(java_node, metadata)

        children = [on_node(child) for child in iterate_java_object(java_node.children())]
        columns = self._extract_columns(java_node, children)

        node = TransformationNode(
            type=self._get_type(),
            metadata=metadata,
            children=children,
            columns=columns
        )
        for column in columns.values():
            column.node_id = node.id

        return node

    def _get_columns(self, java_node: JavaObject) -> JavaObject:
        return java_node.output()

    def _extract_columns(self, java_node: JavaObject, children: List[TransformationNode]) -> Dict[int, TransformationColumn]:
        columns = {}
        for java_column in iterate_java_object(self._get_columns(java_node)):
            column_id = java_column.exprId().id()
            parsed_column = self._extract_column(java_column, children)
            columns[column_id] = parsed_column
        return columns

    def _extract_column(self, java_column: JavaObject, children: List[TransformationNode]) -> TransformationColumn:
        if java_column.nodeName() not in ["Alias", "AttributeReference"]:
            raise NotImplementedError(f"Column type not supported: {java_column.nodeName()}")

        column_name = java_column.name()
        column_id = java_column.exprId().id()
        data_type = java_column.dataType().simpleString()
        reference_ids = [reference.exprId().id() for reference in iterate_java_object(java_column.references())]
        tree_string = java_column.treeString()

        # For each referenced column, search it in the children in order to add a link
        links = []
        for ref_id in reference_ids:
            for child in children:
                if ref_id in child.columns:
                    links.append(child.columns[ref_id])

        return TransformationColumn(
            name=column_name,
            id=column_id,
            type=data_type,
            node_id=None,
            links=links,
            tree_string=tree_string
        )

    def _extract_common_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["schema_string"] = node.schemaString()

    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        pass

    def _get_type(self) -> TransformationType:
        raise NotImplementedError("Abstract method")

    def _expected_number_of_nodes(self) -> Optional[int]:
        raise NotImplementedError("Abstract method")



from setuptools import setup, find_packages

# use repository README as long description
from pathlib import Path
this_directory = Path(__file__).parent
README = (this_directory / "README.md").read_text()


setup(
    name="spark-board",
    version="1.0.0",
    author="Axel Lijdens, Ezequiel Werner",
    url="https://github.com/alijdens/spark-board",
    description="Interactive visualization of Spark jobs",
    long_description_content_type="text/markdown",
    long_description=README,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">3.6",
    install_requires=[
        "pyspark>3",
    ],
    packages=[package for package in find_packages() if package.startswith("spark_board")],
    # this is required to include the static web files that show the spark-board UI in the
    # final package. This directory is created during the deployment process and contains
    # the compiled spark-board-ui react application
    package_data={
        "spark_board": ["ui/**/*"],
    },
)



import os
import shutil
import json
import dataclasses

from spark_board.plan_extractor.errors import RTFMException

from .import env
from .default_settings import DefaultSettings as DefaultSettings  # explicit re-export for mypy
from .plan_extractor import dag
from .plan_extractor.dag_builder import build_dag
from .plan_extractor.dag_simplification import default_simplifiers
from .plan_extractor.dag_simplification.dag_simplifier import DagSimplifier
from .plan_extractor.transformations_dag import JoinCondition, TransformationColumn, TransformationNode, TransformationType
from .plan_extractor.transformation_node_builders.repository import create_default, create_strict, Repository as TnbRepository

from pyspark.sql import DataFrame
from typing import Dict, Any, List, Tuple, Optional


# string template to build the JS file containing the graph definition as nodes
# and links objects required by the framework that will draw them
MODEL_FILE_TEMPLATE = """
const model_staticSettings = {static_settings};

const model_defaultSettings = {settings};

const model_initialNodes = {nodes};

const model_initialEdges = {links};
"""

class Encoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if type(o) is JoinCondition:
            return dataclasses.asdict(o)
        return json.JSONEncoder.default(self, o)    


def _default_tnb_repository() -> TnbRepository:
    return create_default() if env.allow_unknown_transformations() else create_strict()


def dump_dataframe(
        df: DataFrame,
        output_dir: str,
        overwrite: bool,
        default_settings: DefaultSettings,
        dag_simplifiers: List[DagSimplifier] = default_simplifiers(),
        tnb_repository: TnbRepository = _default_tnb_repository(),
        source_code_link: Optional[str] = None,
    ) -> None:
    """
    Create a visual representation of the given `dag` in HTML. The HTML
    files will be saved in the `output_dir` directory. 
    If `overwrite` is True, the output directory will be deleted if it already exists.
    The `dag_simplifiers` are a series of heuristics that will be applied to
    simplify the resulting DAG. spark_board implements some default ones, but they're 
    configurable.
    'tnb_repository' contains the set of known TransformationNodeBuilder instances. The
    user could add more by configuring it.
    `source_code_link` is the link to the source code of the DataFrame. This
    can be a link to a repository, for example. It will appear in the
    generated site so users can go directly to the code.
    """

    tree = build_dag(df=df, dag_simplifiers=dag_simplifiers, tnb_repository=tnb_repository)
    nodes, links = get_nodes_and_links(tree)

    model_file = MODEL_FILE_TEMPLATE.format(
        nodes=json.dumps(nodes, indent=4, cls=Encoder),
        links=json.dumps(links, indent=4),
        settings=json.dumps(dataclasses.asdict(default_settings), indent=4),
        static_settings=json.dumps({"sourceCodeLink": source_code_link}, indent=4),
    )

    if overwrite and os.path.exists(output_dir):
        shutil.rmtree(output_dir)

    try:
        # copy the ui directory to the output directory
        here = os.path.dirname(__file__)
        shutil.copytree(os.path.join(here, "ui"), output_dir)
    except FileNotFoundError:
        # this shouldn't happen in production (installing from PyPI) because the files are
        # included in the package, so it happened during development or we messed up with packaging
        raise RTFMException(
            "The static UI files were not found. If you are running spark-board from source, "
            "make sure you compile it and copy the files there.",
            link="https://github.com/alijdens/spark-board/tree/main/spark-board-ui#test-with-spark-board"
        )

    # create the output file with the model data
    with open(os.path.join(output_dir, "model.js"), "w") as fp:
        fp.write(model_file)


def _get_column_id(column: TransformationColumn) -> str:
    return f"{str(column.node_id)}->{str(column.id)}"


def _column_as_dict(column: TransformationColumn) -> Dict[str, object]:
    return {
        "id": _get_column_id(column),
        "type": "column",
        "parentNode": str(column.node_id),
        "expandParent": True,
        "data": {
            "id": column.id,
            "name": column.name,
            "type": column.type,
            "tree_string": column.tree_string,
            "linked_columns": [_get_column_id(link) for link in column.links]
        },
    } 


def _transformation_as_dict(node: TransformationNode) -> Dict[str, object]:
    # map the Node.type to the type required by the HTML DAG renderer
    node_type_map = {
        TransformationType.Alias: 'Alias',
        TransformationType.Project: 'Project',
        TransformationType.Filter: 'Filter',
        TransformationType.LogicalRDD: 'DataSource',
        TransformationType.Relation: 'DataSource',
        TransformationType.Generate: 'Transform',
        TransformationType.Aggregate: 'Group',
        TransformationType.Join: 'Join',
        TransformationType.Sort: 'Sort',
        TransformationType.Window: 'Window',
        TransformationType.Union: 'Union',
        TransformationType.Limit: 'Limit',
        TransformationType.Repartition: 'Repartition',
        TransformationType.Deduplicate: 'Deduplicate',
        TransformationType.Except: 'Except',
        TransformationType.Intersect: 'Intersect',
        TransformationType.Sample: 'Sample',
        TransformationType.Expand: 'Expand',
        TransformationType.Unknown: 'Unknown',
    }

    return {
        # each node must have a unique ID
        "id": str(node.id),
        # the type of the node for react-flow to use the correct component
        "type": "transformation",
        # data passed to the node upon creation
        "data": {
            # store the transformation type
            "type": node_type_map[node.type],
            # label to display in the node
            "label": node_type_map[node.type],
            # Metadata depending on node type
            "metadata": node.metadata,
            "columns": [_get_column_id(column) for column in node.columns.values()],
        },
    }


def _transformation_link_as_dict(data: dag.DFSNodeData, node: TransformationNode) -> Dict[str, object]:
    return {
        "id": f"{data.parent_id}-{node.id}",
        "source": str(data.parent_id),
        "target": str(node.id),
        "data": {},
    }


def _column_link_as_dict(source_column: TransformationColumn, target_column: TransformationColumn) -> Dict[str, Any]:
    source = f"{str(source_column.node_id)}->{str(source_column.id)}"
    target = f"{str(target_column.node_id)}->{str(target_column.id)}"
    return {
        "id": f"({source})-({target})",
        "source": source,
        "target": target,
        "animated": True,
        "data": {
            "is_column_link": True,
        },
    }


def get_nodes_and_links(tree: TransformationNode) -> Tuple[List[Any], List[Any]]:
    """Convert the tree into a list of nodes and links for the HTML DAG."""

    transformation_nodes, transformation_links, column_nodes, column_links = [], [], [], []
    for data, node in dag.dfs(tree):
        transformation_nodes.append(_transformation_as_dict(node))
        # if the parent is None, it means we are at the root of the tree
        if data.parent_id is not None:
            transformation_links.append(_transformation_link_as_dict(data, node))

        for column in node.columns.values():
            column_nodes.append(_column_as_dict(column))
            for link in column.links:
                column_links.append(_column_link_as_dict(column, link))

    return transformation_nodes + column_nodes, transformation_links + column_links



from ..transformations_dag import TransformationNode


class DagSimplifier(object):
    """
    The DAG Simplifiers detect certain pattern in a DAG, and map them to
    a simplified version of it.
    For example, four filters in a row could be simplified as one filter
    with four conditions.
    """

    def can_apply(self, dag: TransformationNode) -> bool:
        """Determines whether a pattern can be simplified by this simplifier or not"""
        raise NotImplementedError("Abstract method")
    
    def apply(self, original: TransformationNode) -> TransformationNode:
        """Returns a simplified version of the original DAG"""
        raise NotImplementedError("Abstract method")



from py4j.java_gateway import JavaObject
from typing import List, Optional

from ..transformations_dag import Metadata, TransformationType
from .transformation_node_builder import TransformationNodeBuilder

from ..py4j_utils import iterate_java_object


class ExpandNodeBuilder(TransformationNodeBuilder):
    def _extract_metadata(self, node: JavaObject, metadata: Metadata) -> None:
        metadata["projections"] = [self._proj(proj) for proj in iterate_java_object(node.projections())]

    def _proj(self, proj: List[JavaObject]) -> List[str]:
        return [col.sql() for col in iterate_java_object(proj)]

    def _get_type(self) -> TransformationType:
        return TransformationType.Expand

    def _expected_number_of_nodes(self) -> Optional[int]:
        return 1
